# Created: 10/19/2019
# Author:  Emiliano Jordan,
# Project: sejings

__author__ = 'Emiliano Jordan'
__author_email__ = 'emiliano.jordan+sejings@protonmail.com'
__description__ = 'A simple package for rapidly developing system sejings.'
__licence__ = 'Apache License 2.0'
__name__ = 'sejings'
__url__ = 'https://github.com/EmilianoJordan/Sejings'
__version__ = '0.0.5'

__all__ = [
    '__author__', '__author_email__', '__description__', '__licence__',
    '__name__', '__url__', '__version__'
]


# Created: 10/12/2019
# Author:  Emiliano Jordan,
# Project: sejings
import copy
import json
import sys
from pathlib import WindowsPath, Path, PosixPath


def _in_doctest():
    """
    Thanks! https://stackoverflow.com/questions/8116118/
    how-to-determine-whether-code-is-running-in-a-doctest
    """
    if '_pytest.doctest' in sys.modules:
        return True

    if hasattr(sys.modules['__main__'], '_SpoofOut'):
        return True

    if sys.modules['__main__'].__dict__.get('__file__', '').endswith('/pytest'):
        return True

    return False


def blank(string: str):
    return string


def decode_bool(string: str):
    if string.lower() == 'false':
        return False
    if string.lower() == 'true':
        return True


sejings_type_mapping = {
    int: (int, blank),
    float: (float, blank),
    dict: (json.loads, json.dumps),
    list: (json.loads, json.dumps),
    bool: (decode_bool, blank),
    WindowsPath: (Path, blank),
    PosixPath: (Path, blank),
    Path: (Path, blank),
}


class Sejings:

    def __init__(self, value=(..., blank, blank)):

        self._set_value(value)

        super().__setattr__('_children', dict())

    def __call__(self, value=(..., blank, blank), encode=False):

        if isinstance(value, tuple) and value[0] is ...:
            if encode:
                return self._encoder(self._val)
            return self._val

        self._set_value(value)

    def __getattr__(self, item):

        if item == '__wrapped__' and _in_doctest():
            raise AttributeError

        setattr(self, item, Sejings())
        return super().__getattribute__(item)

    def __setattr__(self, key, value):

        if key == '_val':
            super().__setattr__(key, value)
            return

        if isinstance(value, Sejings):
            super().__setattr__(key, value)
            self._children[key] = value
            return

        try:
            obj = super().__getattribute__(key)
            obj(value)
        except AttributeError:
            new_sejings = Sejings(value)
            super().__setattr__(key, new_sejings)
            self._children[key] = new_sejings
            return

    def __len__(self):
        return len(self._children)

    def __copy__(self):

        cls = self.__class__

        new = cls((self._val, self._decoder, self._encoder))

        iterator: dict = copy.copy(self._children)

        for attr, val in iterator.items():
            if isinstance(val, cls):
                setattr(new, attr, val.__copy__())
            else:
                setattr(new, attr, val)

        return new

    def __deepcopy__(self, memodict):

        new_val = copy.deepcopy(self._val, memodict)

        new = self.__class__((new_val, self._decoder, self._encoder))

        iterator: dict = copy.copy(self._children)

        for attr, val in iterator.items():

            if id(val) in memodict:
                setattr(new, attr, memodict[id(val)])

            elif hasattr(val, '__deepcopy__'):
                setattr(new, attr, val.__deepcopy__(memodict))

            elif hasattr(val, 'copy') and callable(getattr(val, 'copy')):

                attr_copy = val.copy()

                if attr is not attr_copy:
                    memodict[id(val)] = attr_copy

                setattr(new, attr, attr_copy)

            elif hasattr(val, '__copy__'):

                attr_copy = val.__copy__()

                if attr is not attr_copy:
                    memodict[id(val)] = attr_copy

                setattr(new, attr, attr_copy)

            else:

                setattr(new, attr, val)

        return new

    def _set_value(self, value):

        if isinstance(value, tuple) \
                and len(value) == 3 \
                and callable(value[1]) \
                and callable(value[2]):
            value, decoder, encoder = value

        elif isinstance(value, tuple) \
                and len(value) == 2 \
                and callable(value[1]):

            value, decoder = value
            encoder = blank

        else:

            try:

                decoder, encoder = sejings_type_mapping[type(value)]

            except KeyError:

                for key, val in sejings_type_mapping.items():
                    if isinstance(value, key):
                        encoder, decoder = val
                        break
                else:
                    encoder, decoder = blank, blank

        super().__setattr__('_val', value)
        super().__setattr__('_decoder', decoder)
        super().__setattr__('_encoder', encoder)

sejings = Sejings()



# Created: 10/12/2019
# Author:  Emiliano Jordan
# Project: sejings

from functools import wraps
from inspect import getfullargspec

from .core import Sejings


def extract_sejings(*ignore_args):
    if (len(ignore_args) == 1
            and callable(ignore_args[0])
            and not isinstance(ignore_args[0], str)):
        # The user function was passed in as args.
        user_function, ignore_args = ignore_args[0], []

        return _extract_sejings_wrapper(user_function, ignore_args)

    def decorator(func):
        return _extract_sejings_wrapper(func, *ignore_args)

    return decorator


def _extract_sejings_wrapper(func, *ignore_args):

    @wraps(func)
    def wrapper(*args, **kwargs):
        # ##### Parse Function Signature for kwarg Sejings ########## #
        # Get the argument specification for the function. This
        # comes with one caveat, if a function has a decorator with
        # a different signature (think functools.lru_cache) then
        # getfullargspec() returns the signature of the decorator
        # not the function. But! in this case func is a partial and
        # has a __wrapped__ attribute that we can use to extract
        # the desired function.

        func_for_argspec = func

        while hasattr(func_for_argspec, '__wrapped__'):
            func_for_argspec = func.__wrapped__

        argspec = getfullargspec(func_for_argspec)

        # Iterate through args to evaluate any Sejings instances
        list_args = list(args)
        for i, (name, val) in enumerate(zip(argspec.args, args)):
            if isinstance(val, Sejings) and name not in ignore_args:
                list_args[i] = val()
        args = tuple(list_args)

        # Iterate through kwargs to evaluate any Sejings instance
        for name, val in kwargs.items():
            if isinstance(val, Sejings) and name not in ignore_args:
                kwargs[name] = val()

        # ####### Get names and default values for all kwargs ######## #
        if argspec.varargs is None and argspec.defaults is not None:
            # The signature has args and kwargs. This is a very typical
            # signature, yet the proper extraction should not be
            # trivialized as a function call could be passing kwargs
            # as positional args.
            defaults_offset = len(args) \
                              + len(argspec.defaults) \
                              - len(argspec.args)

            args_offset = defaults_offset - len(argspec.defaults)

            kwarg_names = argspec.args[args_offset:]
            defaults = argspec.defaults[defaults_offset:]

        elif argspec.defaults is None and argspec.kwonlydefaults is not None:
            # The signature has an Arbitrary Argument List
            # (think *args) and kwargs that follow. This means that the
            # kwargs names are now most easily accessible in kwonlyargs.
            # And the corresponding values are stored in
            # kwonlydefaults dict and not defaults.
            kwarg_names = argspec.kwonlyargs
            defaults = [v for k, v in argspec.kwonlydefaults.items()]

        else:
            # The signature has only args or no args at all.
            # AKA there's no edge cases here! :)
            return func(*args, **kwargs)

        for name, val in zip(kwarg_names, defaults):
            # Evaluate any Sejings instance that is in a kwarg and
            # hasn't been already defined by the function call.
            if name not in kwargs and isinstance(val, Sejings) and name not in ignore_args:
                kwargs[name] = val()

        return func(*args, **kwargs)

    return wrapper



# Created: 12/29/2019
# Author:  Emiliano Jordan,
# Project: sejings
import configparser
import copy
import json
from collections import defaultdict
from pathlib import Path
from typing import Optional, Union, List

from .core import Sejings


def to_configparser(
        sejing,
        config: Optional[configparser.ConfigParser] = None,
        label: str = 'Sejings',
) -> configparser.ConfigParser:
    if config is None:
        config = configparser.ConfigParser()

    config[label] = to_dict(sejing, encode=True)

    return config


def to_dict(sejing, storage_dict=None, key='', encode=False) -> dict:
    if storage_dict is None:
        storage_dict = dict()

    if key and key[0] == '.':
        key = key[1:]

    if sejing() is not ...:
        storage_dict[key] = copy.deepcopy(sejing(encode=encode))

    for iter_key, val in sejing._children.items():
        to_dict(
            sejing=val,
            storage_dict=storage_dict,
            key=f'{key}.{iter_key}',
            encode=encode
        )

    return storage_dict


def to_json(sejing) -> str:
    return json.dumps(to_dict(sejing))


def to_config_file(
        sejing,
        file: Union[str, Path],
        mode: str = 'w',
        label='Sejings',
):
    config = to_configparser(sejing, label=label)

    with open(file, mode=mode) as f:
        config.write(f)


def update_from_dict(sejing, storage_dict: dict):
    sorting_dict = defaultdict(dict)

    for key, val in storage_dict.items():

        if '.' not in key:

            val = copy.deepcopy(val)

            if key not in sejing.__dict__:
                setattr(sejing, key, Sejings(val))
            else:
                obj = getattr(sejing, key)
                obj(obj._decoder(val))

            continue

        attr_parts = key.split('.')
        first_attr, attr_string = attr_parts[0], '.'.join(attr_parts[1:])

        sorting_dict[first_attr][attr_string] = val

    for key, val in sorting_dict.items():

        if key not in sejing.__dict__:
            setattr(sejing, key, Sejings())

        update_from_dict(getattr(sejing, key), val)


def update_from_config_file(
        sejings: Sejings,
        file: Union[str, Path],
        labels: Union[List[str], str, None] = None,
):
    config = configparser.ConfigParser()

    config.read(file)

    if labels is None:
        labels = list(config.keys())
    if not isinstance(labels, list):
        labels = [labels]

    iterator = {
        key: val
        for key, val
        in config.items()
        if key in labels
    }

    build_dict = dict()

    for key, val in iterator.items():
        build_dict = {**build_dict, **val}

    if len(build_dict) == 0:
        raise LookupError(f'label does not exist in file: {file}')

    update_from_dict(sejing=sejings, storage_dict=build_dict)



# Created: 10/18/2019
# Author:  Emiliano Jordan,
# Project: sejings
import os
from setuptools import setup, find_packages


def read(*file):
    with open(os.path.join(os.path.dirname(__file__), *file)) as f:
        return f.read()


about = {}
exec(read('sejings', '__version__.py'), about)

setup(
    author=about['__author__'],
    author_email=about['__author_email__'],
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Utilities',
    ],
    description=about['__description__'],
    license=about['__licence__'],
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    name=about['__name__'],
    packages=find_packages(),
    project_urls={
        'Source': 'https://github.com/EmilianoJordan/Sejings/',
        'Tracker': 'https://github.com/EmilianoJordan/Sejings/issues',
    },
    tests_require=[
        'pytest>=3.5.*',
    ],
    url=about['__url__'],
    version=about['__version__'],
)
